import json
import time
import uuid
import os
import errno
import threading

import yaml

from .exceptions import *
from .base_connection import RetryException
from .connections import connections


CONFIG_FILE = '~/.channelpy.yml'


class Queue(object):

    DEFAULT_RETRY_TIMEOUT = 10

    def __init__(self, name, connection=None, retry_timeout=None):
        self.name = name
        self.connection = connection
        self.retry_timeout = retry_timeout or self.DEFAULT_RETRY_TIMEOUT
        self._try_until_timeout(self._reconnect,
                                timeout=self.retry_timeout)()

    def _reconnect(self):
        self.connection.connect()
        self._queue = self.connection.create_queue(self.name)
        self._event_queue = self.connection.create_local_queue()
        self._pubsub = self.connection.create_pubsub(self.name)
        self.connection.subscribe(self._event_queue, self._pubsub)

    def close(self):
        """Close this instance of the channel."""

        def _close(this):
            this.connection.close()

        t = threading.Thread(target=_close, args=(self,))
        t.start()

    def event(self, ev):
        """Publish an event."""

        self.connection.publish(ev, self._pubsub)

    def delete(self):
        """Delete the queue completely."""

        def _delete(this):
            this.connection.delete_queue(this._queue)
            this.connection.delete_pubsub(this._pubsub)
            this.close()

        t = threading.Thread(target=_delete, args=(self,))
        t.start()

    def _check_for_events(self):
        ev = self.connection.get(self._event_queue)
        if ev is None:
            return
        obj = json.loads(ev.decode('utf-8'))
        if obj == 'close':
            raise ChannelCloseAllException()
        else:
            raise ChannelEventException(obj)

    def _try_until_timeout(self, f, timeout=10, sleep=0.01):
        _f = self.connection.retrying(f)

        def wrapper(*args, **kwargs):
            start = time.time()
            while True:
                try:
                    return _f(*args, **kwargs)
                except RetryException:
                    if time.time() - start > timeout:
                        raise ChannelTimeoutException()
                    time.sleep(sleep)

        return wrapper

    def _retrying(self, f):
        _f = self.connection.retrying(f)

        def wrapper(*args, **kwargs):
            try:
                return _f(*args, **kwargs)
            except RetryException:
                self._try_until_timeout(self._reconnect,
                                        timeout=self.retry_timeout)()
                return f(*args, **kwargs)

        return wrapper

    def _get(self):
        self._check_for_events()
        return self.connection.get(self._queue)

    def get(self):
        return self._retrying(self._get)()

    def _put(self, msg):
        self._check_for_events()
        self.connection.put(msg, self._queue)

    def put(self, msg):
        self._retrying(self._put)(msg)


class ChannelEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, Channel):
            return obj.to_json()
        return super(ChannelEncoder, self).default(obj)


def as_channel(dct):
    if '__channel__' in dct:
        return Channel.from_json(dct)

    return dct


def checking_events(f):

    def wrapper(self, *args, **kwargs):
        try:
            return f(self, *args, **kwargs)
        except ChannelCloseAllException:
            self.close()
            raise ChannelClosedException()

    return wrapper


class Channel(object):

    POLL_FREQUENCY = 0.1  # seconds

    def __init__(self, name=None, persist=True,
                 connection_type=None,
                 retry_timeout=None,
                 **kwargs):
        """
        :type name: str
        :type persist: bool
        :type connection_type: AbstractConnection
        :type kwargs: Dict
        """
        self.name = name or uuid.uuid4().hex
        self.connection_type = None
        self.retry_timeout = retry_timeout
        self.connection_args = {}

        # try first to read config from file
        self._try_config_from_file()
        # if connection type is still empty, use parameter
        if connection_type is not None:
            self.connection_type = connection_type
        if self.connection_type is None:
            raise ChannelInitConnectionException(
                'no connection type found')

        self.connection_args.update(kwargs)
        self.connection = self.connection_type(**self.connection_args)
        self._persist = persist
        self._queue = Queue(self.name, self.connection,
                            retry_timeout=self.retry_timeout)

    def to_json(self):
        return {
            '__channel__': True,
            'name': self.name,
            'connection_type': self.connection_type.__name__,
            'retry_timeout': self.retry_timeout,
            'connection_args': self.connection_args
        }

    @classmethod
    def from_json(cls, obj):
        class_name = obj['connection_type']
        conn_cls = connections[class_name]
        return cls(name=obj['name'],
                   connection_type=conn_cls,
                   retry_timeout=obj['retry_timeout'],
                   **obj['connection_args'])

    def _try_config_from_file(self):
        try:
            with open(os.path.expanduser(CONFIG_FILE)) as config:
                cfg = yaml.load(config)
        except IOError as exc:
            if exc.errno == errno.ENOENT:
                return
            raise
        self.connection_type = connections[cfg.get('connection', None)]
        self.connection_args.update(cfg.get('arguments', {}))
        self.retry_timeout = cfg.get('retry_timeout', None)
        self.POLL_FREQUENCY = float(cfg.get('poll_frequency',
                                            self.POLL_FREQUENCY))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_type, exc_val, exc_tb
        if not self._persist:
            self.delete()
        else:
            self.close()

    def clone(self, name=None, persist=True, retry_timeout=None):
        return Channel(name=name,
                       persist=persist,
                       connection_type=self.connection_type,
                       retry_timeout=retry_timeout,
                       **self.connection_args)

    def dup(self):
        return self.clone(name=self.name, persist=False,
                          retry_timeout=self.retry_timeout)

    @checking_events
    def get(self, timeout=float('inf')):
        start = time.time()
        while True:
            if self._queue is None:
                raise ChannelClosedException()
            msg = self._queue.get()
            if msg is not None:
                return self._process(msg.decode('utf-8'))
            if time.time() - start > timeout:
                raise ChannelTimeoutException()
            time.sleep(self.POLL_FREQUENCY)

    @staticmethod
    def _process(msg):
        return json.loads(msg, object_hook=as_channel)

    @checking_events
    def put(self, value):
        if self._queue is None:
            raise ChannelClosedException()
        self._queue.put(
            json.dumps(value, cls=ChannelEncoder).encode('utf-8'))

    def close(self):
        if self._queue is None:
            raise ChannelClosedException()
        self._queue.close()
        self._queue = None

    def delete(self):
        if self._queue is None:
            raise ChannelClosedException()
        self._queue.delete()
        self.close()

    def close_all(self):
        if self._queue is None:
            raise ChannelClosedException()
        self._queue.event(
            json.dumps('close').encode('utf-8'))
        self.close()

    def event(self, obj):
        if self._queue is None:
            raise ChannelClosedException()
        self._queue.event(json.dumps(obj).encode('utf-8'))

    def put_sync(self, value, timeout=float('inf')):
        """Synchronous put.

        Wraps the object ``value`` in the form::

            {"value": value, "reply_to": ch}

        and waits for a response in ``ch``.

        """
        with Channel(connection_type=self.connection_type,
                     persist=False,
                     **self.connection_args) as ch:
            self.put({
                'value': value,
                'reply_to': ch
            })
            return ch.get(timeout=timeout)
